#include "HumanTracker.h"

#include "ArmarXCore/core/exceptions/local/ExpressionException.h"

#include "armarx/navigation/human/types.h"
#include <armarx/navigation/conversions/eigen.h>
#include <range/v3/range/conversion.hpp>
#include <range/v3/view/transform.hpp>

namespace armarx::navigation::human
{
    void
    HumanTracker::update(const Measurements& measurements)
    {
        // iterate over all existing tracked humans
        for (auto it = trackedHumans.begin(); it != trackedHumans.end();)
        {
            auto& human = *it;
            // when the tracked human recieved no new measurement for too long, remove it
            if ((measurements.detectionTime - human.humanFilter.get().detectionTime) >=
                parameters.maxTrackingAge)
            {
                it = trackedHumans.erase(it);
            }
            // otherwise the tracked human is prepared for association at the current point in time
            else
            {
                human.associated = false;
                human.humanFilter.propagation(measurements.detectionTime);
                it++;
            }
        }

        // calculate the poses according to the new received measurements
        std::vector<DetectedHuman> newPoses =
            measurements.humanPoses |
            ranges::views::transform(
                [measurements, this](const armem::human::HumanPose& humanPose) -> DetectedHuman {
                    return convertHumanPoseToDetectedHuman(measurements.detectionTime, humanPose);
                }) |
            ranges::to_vector;

        // associate the new poses from the measurement with the tracked humans
        associateHumans(newPoses);

        // add all not associated poses as new tracked humans
        for (const auto& detectedHuman : newPoses)
        {
            if (not detectedHuman.associated)
            {
                //add new tracked human to list of tracked humans
                trackedHumans.push_back(TrackedHuman{
                    .humanFilter = HumanFilter{detectedHuman.pose, detectedHuman.detectionTime},
                    .trackingId = detectedHuman.trackingId,
                    .associated = true});
            }
        }
    }

    std::vector<human::Human>
    HumanTracker::getTrackedHumans() const
    {
        return trackedHumans |
               ranges::views::transform([](const TrackedHuman& h) -> human::Human
                                        { return h.humanFilter.get(); }) |
               ranges::to_vector;
    }

    void
    HumanTracker::reset()
    {
        trackedHumans.clear();
    }


    struct PosDistance
    {
        HumanTracker::TrackedHuman* oldHuman;
        HumanTracker::DetectedHuman* newHuman;
        float distance;
    };

    std::vector<PosDistance>
    getSortedDistances(std::vector<HumanTracker::TrackedHuman>& oldHumans,
                       std::vector<HumanTracker::DetectedHuman>& newHumans)
    {
        std::vector<PosDistance> posDistances;

        for (auto& oldHuman : oldHumans)
        {
            if (oldHuman.associated)
            {
                continue;
            }
            for (auto& newHuman : newHumans)
            {
                if (newHuman.associated)
                {
                    continue;
                }
                // calculate distance between every possible combination of tracked and detected
                // humans where none of them was associated
                posDistances.push_back(
                    {&oldHuman,
                     &newHuman,
                     (newHuman.pose.translation() - oldHuman.humanFilter.get().pose.translation())
                         .norm()});
            }
        }

        // sort the distances ascending by their numeric value
        std::sort(posDistances.begin(),
                  posDistances.end(),
                  [](const PosDistance& a, const PosDistance& b) -> bool
                  { return a.distance < b.distance; });

        return posDistances;
    }


    HumanTracker::DetectedHuman
    HumanTracker::convertHumanPoseToDetectedHuman(const DateTime& time,
                                                  const armem::human::HumanPose& humanPose)
    {
        const std::map<std::string, armem::human::PoseKeypoint>& keypoints = humanPose.keypoints;
        ARMARX_CHECK_NOT_EMPTY(keypoints);

        // calculate the arithmetic mean of all keypoint positions
        Eigen::Vector3f centerPos;
        int size = 0;
        for (const auto& [_, v] : keypoints)
        {
            if (v.positionGlobal.has_value())
            {
                centerPos += v.positionGlobal.value().toEigen();
                size++;
            }
        }
        centerPos /= size;

        // calculate the yaw of the specified keypoint representing the orientation if it exists
        double yaw = 0;
        if (humanPose.keypoints.count(parameters.rotationKeypoint) > 0)
        {
            const armem::human::PoseKeypoint& refKeypoint =
                humanPose.keypoints.at(parameters.rotationKeypoint);
            ARMARX_CHECK(refKeypoint.orientationGlobal.has_value());
            Eigen::Quaternionf qhead = refKeypoint.orientationGlobal->toEigenQuaternion();
            Eigen::Vector3f vec(1, 0, 0);
            Eigen::Vector3f rot3 = qhead._transformVector(vec);
            Eigen::Vector2f rot2(rot3.x(), rot3.y());
            if (rot2.norm() != 0)
            {
                // calculate angle of rot2
                yaw = atan2(rot2.y(), rot2.x());
            }
            //old version using euler angles:
            //yaw = humanPose //from all human pose keypoints
            //          .keypoints
            //          .at("HEAD") //find the keypoint representing the head
            //          .orientationGlobal //get its global orientation
            //          ->toEigen()
            //          .eulerAngles(2, 1, 0)[0]; //and extract the yaw (rotation around z axis in
            //                                                                  global coordinates)
        }

        // create the new pose with the calculated position and yaw
        core::Pose2D pose = core::Pose2D::Identity();
        pose.translation() = conv::to2D(centerPos);
        pose.linear() = Eigen::Rotation2Df(yaw).toRotationMatrix();

        return {pose, humanPose.humanTrackingId, time, false};
    }

    void
    HumanTracker::associateHumans(std::vector<DetectedHuman>& detectedHumans)
    {
        // first, associate humans by their tracking id
        for (auto& oldHuman : trackedHumans)
        {
            if (oldHuman.associated || not oldHuman.trackingId)
            {
                continue;
            }
            for (auto& newHuman : detectedHumans)
            {
                if (newHuman.associated || not newHuman.trackingId)
                {
                    continue;
                }
                // check for every possible combination of old and new humans (that were not already
                // associated and have a tracking id) if their tracking id is the same
                if (oldHuman.trackingId.value() == newHuman.trackingId.value())
                {
                    associate(&oldHuman, &newHuman);
                }
            }
        }

        // second, associate leftover humans by their distances
        const auto sortedDistances = getSortedDistances(trackedHumans, detectedHumans);

        for (auto& posDistance : sortedDistances)
        {
            if (posDistance.distance > parameters.maxAssociationDistance)
            {
                break;
            }
            if (posDistance.oldHuman->associated || posDistance.newHuman->associated)
            {
                continue;
            }
            // associate the pair with the currently smallest distance between non-associated humans
            associate(posDistance.oldHuman, posDistance.newHuman);
        }
    }

    void
    HumanTracker::associate(TrackedHuman* trackedHuman, DetectedHuman* detectedHuman)
    {
        ARMARX_CHECK(not trackedHuman->associated);
        ARMARX_CHECK(not detectedHuman->associated);

        trackedHuman->associated = true;
        detectedHuman->associated = true;

        trackedHuman->humanFilter.update(detectedHuman->pose, detectedHuman->detectionTime);
        trackedHuman->trackingId = detectedHuman->trackingId;
    }

} // namespace armarx::navigation::human
